# Functions to manipulate Python binary distribution archives.
#
# Author: Peter Odding <peter.odding@paylogic.eu>
# Last Change: May 15, 2014
# URL: https://github.com/paylogic/pip-accel

"""
Binary distribution archive manipulation
========================================

The functions in this module are used to create, transform and install from
binary distribution archives.
"""

# Standard library modules.
import hashlib
import logging
import os
import os.path
import pipes
import re
import shutil
import subprocess
import tarfile
import tempfile
import time

# External dependencies.
from humanfriendly import Spinner, Timer

# Modules included in our package.
from pip_accel.config import binary_index
from pip_accel.deps import sanity_check_dependencies
from pip_accel.utils import get_python_version

# Initialize a logger for this module.
logger = logging.getLogger(__name__)

def get_binary_dist(package, version, directory, url=None, python='/usr/bin/python', prefix='/usr'):
    """
    Get the cached binary distribution archive that was previously built for
    the given package (name, version) (and optionally URL). If no archive has
    been cached yet, a new binary distribution archive is created and added to
    the cache.

    :param package: The name of the requirement to build.
    :param version: The version of the requirement to build.
    :param directory: The directory where the unpacked sources of the
                      requirement are available.
    :param url: The URL of the requirement, optional. When given this is used
                to generate the filename of the cached binary distribution.
    :param python: The pathname of the Python executable to use to run
                   ``setup.py`` (obviously this should point to a working
                   Python installation).
    :param prefix: The prefix that the original binary distribution was created for.
    :returns: An iterable of tuples with two values each: A
              :py:class:`tarfile.TarInfo` object and a file-like object.
    """
    tag = hashlib.sha1(version + url).hexdigest() if url else version
    cache_file = os.path.join(binary_index, '%s:%s:%s.tar.gz' % (package, tag, get_python_version()))
    if not os.path.isfile(cache_file):
        logger.debug("%s (%s) hasn't been cached yet, doing so now.", package, version)
        # Build the binary distribution.
        try:
            raw_file = build_binary_dist(package, version, directory, python=python)
        except BuildFailed:
            sanity_check_dependencies(package)
            raw_file = build_binary_dist(package, version, directory, python=python)
        # Transform the binary distribution archive into a form that we can re-use.
        transformed_file = '%s.tmp-%i' % (cache_file, os.getpid())
        archive = tarfile.open(transformed_file, 'w:gz')
        for member, from_handle in transform_binary_dist(raw_file, prefix=prefix):
            archive.addfile(member, from_handle)
        archive.close()
        # Try to avoid race conditions between multiple processes by atomically
        # moving the transformed binary distribution into its final place.
        os.rename(transformed_file, cache_file)
        logger.debug("%s (%s) cached as %s.", package, version, cache_file)
    archive = tarfile.open(cache_file, 'r:gz')
    for member in archive.getmembers():
        yield member, archive.extractfile(member.name)
    archive.close()

def build_binary_dist(package, version, directory, python='/usr/bin/python'):
    """
    Convert a single, unpacked source distribution to a binary distribution.
    Raises an exception if it fails to create the binary distribution (probably
    because of missing binary dependencies like system libraries).

    :param package: The name of the requirement to build.
    :param version: The version of the requirement to build.
    :param directory: The directory where the unpacked sources of the
                      requirement are available.
    :param python: The pathname of the Python executable to use to run
                   ``setup.py`` (obviously this should point to a working
                   Python installation).
    :returns: The pathname of the resulting binary distribution (a string).
    """
    build_timer = Timer()
    # Make sure the source distribution contains a setup script.
    setup_script = os.path.join(directory, 'setup.py')
    if not os.path.isfile(setup_script):
        msg = "Directory %s (%s %s) doesn't contain a source distribution!"
        raise InvalidSourceDistribution(msg % (directory, package, version))
    # Let the user know what's going on.
    build_text = "Building binary distribution of %s (%s) .." % (package, version)
    logger.info("%s", build_text)
    # Cleanup previously generated distributions.
    dist_directory = os.path.join(directory, 'dist')
    if os.path.isdir(dist_directory):
        logger.debug("Cleaning up previously generated distributions in %s ..", dist_directory)
        shutil.rmtree(dist_directory)
    # Compose the command line needed to build the binary distribution.
    command_line = '%s setup.py bdist_dumb --format=tar' % pipes.quote(python)
    logger.debug("Executing external command: %s", command_line)
    # Redirect all output of the build to a temporary file.
    fd, temporary_file = tempfile.mkstemp()
    command_line = '%s > "%s" 2>&1' % (command_line, temporary_file)
    try:
        # Start the build.
        build = subprocess.Popen(['sh', '-c', command_line], cwd=directory)
        # Wait for build to finish, provide feedback to the user in the mean time.
        spinner = Spinner(build_text)
        while build.poll() is None:
            spinner.step()
            time.sleep(0.1)
        spinner.clear()
        # Check whether the build succeeded.
        if build.returncode != 0:
            # It it didn't we'll provide the user with some hints as to what went wrong.
            msg = "Failed to build binary distribution of %s (%s)!" % (package, version)
            logger.error("%s", msg)
            with open(temporary_file) as handle:
                logger.info("Build output (will probably provide a hint as to what went wrong):\n%s", handle.read())
            raise BuildFailed(msg)
        # Check if we can find the binary distribution archive.
        filenames = os.listdir(dist_directory)
        if len(filenames) != 1:
            msg = "Build process did not result in one binary distribution! (matches: %s)"
            raise NoBuildOutput(msg % filenames)
        logger.info("Finished building %s (%s) in %s.", package, version, build_timer)
        return os.path.join(dist_directory, filenames[0])
    finally:
        os.unlink(temporary_file)

def transform_binary_dist(archive_path, prefix='/usr'):
    """
    Transform a binary distribution archive created with ``python setup.py
    bdist_dumb --format=tar`` into a form that can be cached for future use.
    This comes down to making the pathnames inside the archive relative to the
    `prefix` that the binary distribution was built for.

    :param archive_path: The pathname of the original binary distribution archive.
    :param prefix: The prefix that the original binary distribution was created for.
    :returns: An iterable of tuples with two values each: A
              :py:class:`tarfile.TarInfo` object and a file-like object.
    """
    # Copy the tar archive file by file so we can rewrite the pathnames.
    logger.debug("Transforming binary distribution: %s.", archive_path)
    logger.debug("Using environment prefix: %s.", prefix)
    archive = tarfile.open(archive_path, 'r')
    for member in archive.getmembers():
        # In my testing the `dumb' tar files created with the `python setup.py
        # bdist' command contain pathnames that are relative to `/' which is
        # kind of awkward: I would like to use os.path.relpath() on them but
        # that won't give the correct result without some preprocessing...
        original_pathname = member.name
        absolute_pathname = re.sub(r'^\./', '/', original_pathname)
        if member.isdev():
            logger.warn("Ignoring device file: %s.", absolute_pathname)
        elif not member.isdir():
            modified_pathname = os.path.relpath(absolute_pathname, prefix)
            if os.path.isabs(modified_pathname):
                logger.warn("Failed to transform pathname in binary distribution to relative path! (original: %r, modified: %r)",
                            original_pathname, modified_pathname)
            else:
                # Rewrite /usr/local to /usr (same goes for all prefixes of course).
                modified_pathname = re.sub('^local/', '', modified_pathname)
                logger.debug("Transformed %r -> %r.", original_pathname, modified_pathname)
                # Get the file data from the input archive.
                handle = archive.extractfile(original_pathname)
                # Yield the pathname, file mode and a handle to the data.
                member.name = modified_pathname
                yield member, handle
    archive.close()

def install_binary_dist(members, prefix, python='/usr/bin/python', enable_workarounds=True):
    """
    Install a binary distribution created with ``python setup.py bdist`` into
    the given prefix (a directory like ``/usr``, ``/usr/local`` or a virtual
    environment).

    :param members: An iterable of tuples with three values each:
                    (pathname, mode, handle).
    :param prefix: The "prefix" under which the requirements should be
                   installed. This will be a pathname like ``/usr``,
                   ``/usr/local`` or the pathname of a virtual environment.
    :param python: The pathname of the Python executable to use in the shebang
                   line of all executable Python scripts inside the binary
                   distribution.
    :param enable_workarounds: Enable workarounds to make the resulting filenames
                               compatible with virtual environments.
    """
    # TODO This is quite slow for modules like Django. Speed it up! Two choices:
    #  1. Run the external tar program to unpack the archive. This will
    #     slightly complicate the fixing up of hashbangs.
    #  2. Using links? The plan: We can maintain a "seed" environment under
    #     $PIP_ACCEL_CACHE and use symbolic and/or hard links to populate other
    #     places based on the "seed" environment.
    for member, from_handle in members:
        pathname = member.name
        # Some binary distributions include C header files (see for example the
        # greenlet package) however the subdirectory of include/ in a virtual
        # environment is a symbolic link to a subdirectory of /usr/include/ so
        # we should never try to install C header files inside the directory
        # pointed to by the symbolic link. Instead we implement the same
        # workaround that pip uses to avoid this problem.
        if enable_workarounds:
            pathname = re.sub('^include/', 'include/site/', pathname)
        pathname = os.path.join(prefix, pathname)
        directory = os.path.dirname(pathname)
        if not os.path.isdir(directory):
            logger.debug("Creating directory: %s ..", directory)
            os.makedirs(directory)
        logger.debug("Creating file: %s ..", pathname)
        with open(pathname, 'wb') as to_handle:
            contents = from_handle.read()
            if contents.startswith(b'#!/'):
                contents = fix_hashbang(python, contents)
            to_handle.write(contents)
        os.chmod(pathname, member.mode)

def fix_hashbang(python, contents):
    """
    Rewrite the hashbang in an executable script so that the Python program
    inside the virtual environment is used instead of a system wide Python.

    :param python: The absolute pathname of the Python program inside the
                   virtual environment.
    :param contents: A string with the contents of the script whose hashbang
                     should be fixed.
    :returns: The modified contents of the script as a string.
    """
    # Separate the first line in the file from the remainder of the contents
    # while preserving the end of line sequence (CR+LF or just an LF) and
    # without having to split all lines in the file (there's no point).
    lines = contents.splitlines()
    hashbang = lines[0]
    # Get the base name of the command in the hashbang and deal with hashbangs
    # like `#!/usr/bin/env python'.
    modified_name = re.sub(b'^env ', b'', os.path.basename(hashbang))
    # Only rewrite hashbangs that actually involve Python.
    if re.match(b'^python(\\d+(\\.\\d+)*)?$', modified_name):
        lines[0] = b'#!' + python.encode('ascii')
        logger.debug("Hashbang %r looks like a Python hashbang! Rewriting it to %r!", hashbang, lines[0])
        contents = b'\n'.join(lines)
    else:
        logger.debug("Warning: Failed to match hashbang: %r.", hashbang)
    return contents

class InvalidSourceDistribution(Exception):
    """
    Raised by :py:func:`build_binary_dist()` when the given directory doesn't
    contain a source distribution.
    """

class BuildFailed(Exception):
    """
    Raised by :py:func:`build_binary_dist()` when a binary distribution build fails.
    """

class NoBuildOutput(Exception):
    """
    Raised by :py:func:`build_binary_dist()` when a binary distribution build
    fails to produce a binary distribution archive.
    """



# Accelerator for pip, the Python package manager.
#
# Author: Peter Odding <peter.odding@paylogic.eu>
# Last Change: May 15, 2014
# URL: https://github.com/paylogic/pip-accel
#
# TODO Permanently store logs in the pip-accel directory (think about log rotation).
# TODO Maybe we should save the output of `python setup.py bdist_dumb` somewhere as well?

"""
The Python module :py:mod:`pip_accel` defines the classes and functions that
implement the functionality of the pip accelerator and the ``pip-accel``
command. Instead of using the ``pip-accel`` command you can also use the pip
accelerator as a Python module. In this case you'll probably want to start by
taking a look at the following functions:

- :py:func:`unpack_source_dists`
- :py:func:`download_source_dists`
- :py:func:`install_requirements`
"""

# Semi-standard module versioning.
__version__ = '0.12.1'

# Standard library modules.
import os
import os.path
import pipes
import shutil
import sys
import tempfile
import textwrap

try:
    # Python 2.x.
    from urllib import unquote
    from urlparse import urlparse
except ImportError:
    # Python 3.x.
    from urllib.parse import unquote
    from urllib.parse import urlparse

# Modules included in our package.
from pip_accel.bdist import get_binary_dist, install_binary_dist
from pip_accel.config import (binary_index, download_cache,
                              index_version_file, source_index)
from pip_accel.req import Requirement
from pip_accel.logger import logger

# External dependencies.
import coloredlogs
from humanfriendly import Timer
from pip import index as pip_index_module
from pip import parseopts
from pip.cmdoptions import requirements as requirements_option
from pip.commands import install as pip_install_module
from pip.commands.install import InstallCommand
from pip.exceptions import DistributionNotFound, InstallationError
from pip.log import logger as pip_logger
from pip.status_codes import SUCCESS

# Find the environment where requirements are to be installed.
ENVIRONMENT = os.path.abspath(os.environ.get('VIRTUAL_ENV', sys.prefix))

# The main loop of pip-accel retries at most this many times to counter pip errors
# due to connectivity issues with PyPI and/or linked distribution websites.
MAX_RETRIES = 10

# The version number of the binary distribution cache format in use. When we
# break backwards compatibility we bump this number so that pip-accel knows it
# should clear the cache before proceeding.
CACHE_FORMAT_REVISION = 4

def main():
    """
    Main logic of the ``pip-accel`` command.
    """
    arguments = sys.argv[1:]
    # If no arguments are given, the help text of pip-accel is printed.
    if not arguments:
        print_usage()
        sys.exit(0)
    # If no install subcommand is given we pass the command line straight
    # to pip without any changes and exit immediately afterwards.
    elif 'install' not in arguments:
        sys.exit(os.spawnvp(os.P_WAIT, 'pip', ['pip'] + arguments))
    # Initialize logging output.
    coloredlogs.install()
    # Increase verbosity based on -v, --verbose options.
    for argument in arguments:
        if argument == '--verbose' or (len(argument) >= 2 and argument[0] ==
                '-' and argument[1] != '-' and 'v' in argument):
            coloredlogs.increase_verbosity()
    # Make sure the prefix is the same as the environment.
    if not os.path.samefile(sys.prefix, ENVIRONMENT):
        logger.error("You are trying to install packages in environment #1 which is different from environment #2 where pip-accel is installed! Please install pip-accel under environment #1 to install packages there.")
        logger.info("Environment #1: %s ($VIRTUAL_ENV)", ENVIRONMENT)
        logger.info("Environment #2: %s (installation prefix)", sys.prefix)
        sys.exit(1)
    main_timer = Timer()
    initialize_directories()
    build_directory = tempfile.mkdtemp()
    # Execute "pip install" in a loop in order to retry after intermittent
    # error responses from servers (which can happen quite frequently).
    try:
        for i in xrange(1, MAX_RETRIES):
            try:
                requirements = unpack_source_dists(arguments, build_directory)
            except DistributionNotFound:
                logger.warn("We don't have all source distributions yet!")
                download_source_dists(arguments, build_directory)
            else:
                install_requirements(requirements)
                logger.info("Done! Took %s to install %i package%s.", main_timer, len(requirements), '' if len(requirements) == 1 else 's')
                return
            logger.warn("pip failed, retrying (%i/%i) ..", i + 1, MAX_RETRIES)
    except InstallationError:
        # Abort early when pip reports installation errors.
        logger.fatal("pip reported unrecoverable installation errors. Please fix and rerun!")
        sys.exit(1)
    finally:
        # Always cleanup temporary build directory.
        shutil.rmtree(build_directory)
    # Abort when after N retries we still failed to download source distributions.
    logger.fatal("External command failed %i times, aborting!" % MAX_RETRIES)
    sys.exit(1)

def print_usage():
    """
    Report the usage of the pip-accel command to the console.
    """
    print(textwrap.dedent("""
        Usage: pip-accel [ARGUMENTS TO PIP]

        The pip-accel program is a wrapper for pip, the Python package manager. It
        accelerates the usage of pip to initialize Python virtual environments given
        one or more requirements files. The pip-accel command supports all subcommands
        and options supported by pip, however it is only useful for the "pip install"
        subcommand.

        For more information please refer to the GitHub project page
        at https://github.com/paylogic/pip-accel
    """).strip())

def clear_build_directory(directory):
    """
    Since pip 1.4 there is a new exception that's raised by pip:
    ``PreviousBuildDirError``. Unfortunately pip-accel apparently triggers the
    worst possible side effect of this new "feature" and the only way to avoid
    it is to start with an empty build directory in every step of pip-accel's
    process (possibly very inefficient).

    :param directory: The build directory to clear.
    """
    logger.debug("Clearing build directory ..")
    if os.path.isdir(directory):
        shutil.rmtree(directory)
    os.makedirs(directory)

def unpack_source_dists(arguments, build_directory):
    """
    Check whether there are local source distributions available for all
    requirements, unpack the source distribution archives and find the names
    and versions of the requirements. By using the ``pip install --no-install``
    command we avoid reimplementing the following pip features:

    - Parsing of ``requirements.txt`` (including recursive parsing)
    - Resolution of possibly conflicting pinned requirements
    - Unpacking source distributions in multiple formats
    - Finding the name & version of a given source distribution

    :param arguments: A list of strings with the command line arguments to be
                      passed to the ``pip`` command.

    :returns: A list of :py:class:`pip_accel.req.Requirement` objects. If
              ``pip`` fails, an exception will be raised by ``pip``.
    """
    unpack_timer = Timer()
    logger.info("Unpacking local source distributions ..")
    clear_build_directory(build_directory)
    try:
        install_custom_package_finder()
        # Execute pip to unpack the source distributions.
        requirement_set = run_pip(arguments + ['--no-install'],
                                  use_remote_index=False,
                                  build_directory=build_directory)
        logger.info("Unpacked local source distributions in %s.", unpack_timer)
        # XXX This feels (looks) like a nasty hack but it prevents an unhandled
        # exception that was introduced in pip-accel==0.11. Please refer to
        # https://github.com/paylogic/pip-accel/issues/24 for gory details.
        filtered_requirements = []
        for requirement in requirement_set.requirements.values():
            if requirement.satisfied_by:
                logger.info("Requirement already satisfied: %s.", requirement)
            else:
                filtered_requirements.append(requirement)
        return sorted([Requirement(r) for r in filtered_requirements],
                      key=lambda r: r.name.lower())
    finally:
        cleanup_custom_package_finder()

def download_source_dists(arguments, build_directory):
    """
    Download missing source distributions.

    :param arguments: A list with the arguments intended for ``pip``.
    """
    download_timer = Timer()
    logger.info("Downloading source distributions ..")
    clear_build_directory(build_directory)
    # Execute pip to download missing source distributions.
    try:
        run_pip(arguments + ['--no-install'], use_remote_index=True, build_directory=build_directory)
        logger.info("Finished downloading source distributions in %s.", download_timer)
    except Exception as e:
        logger.warn("pip raised an exception while downloading source distributions: %s.", e)

def install_requirements(requirements, install_prefix=ENVIRONMENT):
    """
    Manually install all requirements from binary distributions.

    :param requirements: A list of :py:class:`pip_accel.req.Requirement` objects.
    :param install_prefix: The "prefix" under which the requirements should be
                           installed. This will be a pathname like ``/usr``,
                           ``/usr/local`` or the pathname of a virtual
                           environment.
    :returns: ``True`` if it succeeds in installing all requirements from
              binary distribution archives, ``False`` otherwise.
    """
    install_timer = Timer()
    logger.info("Installing from binary distributions ..")
    python = os.path.join(install_prefix, 'bin', 'python')
    pip = os.path.join(install_prefix, 'bin', 'pip')
    for requirement in requirements:
        if os.system('%s uninstall --yes %s >/dev/null 2>&1' % (pipes.quote(pip), pipes.quote(requirement.name))) == 0:
            logger.info("Uninstalled previously installed package %s.", requirement.name)
        members = get_binary_dist(requirement.name, requirement.version,
                                  requirement.source_directory, requirement.url,
                                  prefix=install_prefix, python=python)
        install_binary_dist(members, prefix=install_prefix, python=python)
    logger.info("Finished installing all requirements in %s.", install_timer)
    return True

def run_pip(arguments, use_remote_index, build_directory=None):
    """
    Execute a modified ``pip install`` command. This function assumes that the
    arguments concern a ``pip install`` command (:py:func:`main()` makes sure
    of this).

    :param arguments: A list of strings containing the arguments that will be
                      passed to ``pip``.
    :param use_remote_index: A boolean indicating whether ``pip`` is allowed to
                             contact http://pypi.python.org.
    :returns: A ``RequirementSet`` object created by ``pip``, unless an
              exception is raised by ``pip`` (in which case the exception will
              bubble up).
    """
    command_line = []
    for i, arg in enumerate(arguments):
        if arg == 'install':
            command_line += ['pip'] + arguments[:i+1] + [
                    '--download-cache=%s' % download_cache,
                    '--find-links=file://%s' % source_index]
            if build_directory:
                command_line += ['--build-directory=%s' % build_directory]
            if not use_remote_index:
                command_line += ['--no-index']
            command_line += arguments[i+1:]
            break
    else:
        command_line = ['pip'] + arguments
    logger.info("Executing command: %s", ' '.join(command_line))
    # XXX Nasty hack required for pip 1.4 compatibility (workaround for global state).
    requirements_option.default = []
    cmd_name, options, args, parser = parseopts(command_line[1:])
    pip = CustomInstallCommand(parser)
    exit_status = pip.main(args[1:], options)
    # Make sure the output of pip and pip-accel are not intermingled.
    sys.stdout.flush()
    update_source_dists_index()
    if exit_status == SUCCESS:
        return pip.requirement_set
    else:
        raise pip.intercepted_exception

class CustomInstallCommand(InstallCommand):

    """
    Subclass of :py:class:`pip.commands.install.InstallCommand` that makes it
    easier to run ``pip install`` from Python. Used by the :py:func:`run_pip()`
    function in order to run a ``pip install`` command in the same process,
    without running pip as a subprocess.
    """

    def main(self, *args, **kw):
        """
        ``pip.basecommand.Command.main()`` expects to be executed only once; it
        unconditionally executes ``pip.log.logger.consumers.extend()``. This
        means that when we run ``pip`` more than once we'll cause it to repeat
        its output as many times as we executed a ``pip install`` command. We
        wrap ``main()`` to explicitly reset the list of consumers.
        """
        pip_logger.consumers = []
        return super(CustomInstallCommand, self).main(*args, **kw)


    def run(self, *args, **kw):
        """
        The method ``pip.commands.install.InstallCommand.run()`` returns a
        ``RequirementSet`` object which ``pip-accel`` is interested in, however
        ``pip.basecommand.Command.main()`` (the caller of ``run()``) swallows
        the requirement set (based on my reading of the pip 1.3.x source code).
        We wrap ``run()`` so that we can intercept the requirement set. This is
        a bit sneaky, but I don't fancy reimplementing large parts of
        ``pip.basecommand.Command.main()`` inside ``pip-accel``!
        """
        original_method = super(CustomInstallCommand, self).run
        try:
            self.intercepted_exception = None
            self.requirement_set = original_method(*args, **kw)
            return self.requirement_set
        except (Exception, KeyboardInterrupt) as e:
            self.intercepted_exception = e
            raise

def update_source_dists_index():
    """
    Link newly downloaded source distributions into the local index directory
    using symbolic links.
    """
    link_timer = Timer()
    for download_name in os.listdir(download_cache):
        download_path = os.path.join(download_cache, download_name)
        url = unquote(download_name)
        if not url.endswith('.content-type'):
            components = urlparse(url)
            archive_name = os.path.basename(components.path)
            archive_path = os.path.join(source_index, add_extension(download_path, archive_name))
            if not os.path.isfile(archive_path):
                logger.debug("Linking files:")
                logger.debug(" - Source: %s", download_path)
                logger.debug(" - Target: %s", archive_path)
                os.symlink(download_path, archive_path)
    logger.debug("Updated source index links in %s.", link_timer)

def add_extension(download_path, archive_path):
    """
    Make sure all cached source distributions have the right file extension,
    because not all distribution sites provide URLs with proper filenames in
    them while we really need the proper filenames to build the local source
    index.

    :param download_path: The pathname of the source distribution archive in
                          the download cache.
    :param archive_path: The pathname of the distribution archive in the source
                         index directory.
    :returns: The (possibly modified) pathname of the distribution archive in
              the source index directory.

    Previously this used the ``file`` executable, now it checks the magic file
    headers itself. I could have used any of the numerous ``libmagic`` bindings
    on PyPI, but that would add a binary dependency to ``pip-accel`` and I
    don't want that :-).
    """
    handle = open(download_path, 'rb')
    header = handle.read(2)
    handle.close()
    if header.startswith(b'\x1f\x8b'):
        # The gzip compression header is two bytes: 0x1F, 0x8B.
        if not archive_path.endswith(('.tgz', '.tar.gz')):
            archive_path += '.tar.gz'
    elif header.startswith(b'BZ'):
        # The bzip2 compression header is two bytes: B, Z.
        if not archive_path.endswith('.bz2'):
            archive_path += '.bz2'
    elif header.startswith(b'PK'):
        # According to Wikipedia, ZIP archives don't have an official magic
        # number, but most of the time we'll find two bytes: P, K (for Phil
        # Katz, creator of the format).
        if not archive_path.endswith('.zip'):
            archive_path += '.zip'
    return archive_path

def initialize_directories():
    """
    Create the directories for the download cache, the source index and the
    binary index if any of them don't exist yet and reset the binary index
    when its format changes.
    """
    # Create all required directories on the fly.
    for directory in [download_cache, source_index, binary_index]:
        if not os.path.isdir(directory):
            os.makedirs(directory)
    # When files are removed from pip's download cache, broken symbolic links
    # remain in pip-accel's source index. This results in very confusing error
    # messages. To avoid this we cleanup broken symbolic links.
    for entry in sorted(os.listdir(source_index)):
        pathname = os.path.join(source_index, entry)
        if os.path.islink(pathname) and not os.path.exists(pathname):
            logger.warn("Cleaning up broken symbolic link: %s", pathname)
            os.unlink(pathname)
    # If 1) pip's download cache is full but 2) pip-accel's source index hasn't
    # been initialized yet and 3) all requirements are available in pip's
    # download cache we can waste a lot of time. To avoid this we update the
    # symbolic links in pip-accel's source index before every run.
    update_source_dists_index()
    # Invalidate the binary distribution cache when the
    # format is changed in backwards incompatible ways.
    if os.path.isfile(index_version_file):
        with open(index_version_file) as handle:
            if int(handle.read()) == CACHE_FORMAT_REVISION:
                logger.debug("Binary distribution cache format is compatible.")
                return
    logger.debug("Binary distribution cache format is incompatible; clearing cache ..")
    for entry in sorted(os.listdir(binary_index)):
        pathname = os.path.join(binary_index, entry)
        logger.debug(" - Deleting %s.", pathname)
        os.unlink(pathname)
    with open(index_version_file, 'w') as handle:
        handle.write("%i\n" % CACHE_FORMAT_REVISION)

ORIGINAL_PACKAGE_FINDER = None

def install_custom_package_finder():
    """
    Install :py:class:`CustomPackageFinder` so we can be sure that pip will not
    try to fetch any index pages (i.e. we disable all crawling, which in my
    experience is the slowest operation performed by pip).
    """
    global ORIGINAL_PACKAGE_FINDER
    logger.debug("Installing custom package finder (to force --no-index behavior) ..")
    ORIGINAL_PACKAGE_FINDER = pip_index_module.PackageFinder
    pip_install_module.PackageFinder = CustomPackageFinder

def cleanup_custom_package_finder():
    """
    Clean up the monkey patch applied by :py:func:`install_custom_package_finder()`.
    Use a try/finally block to ensure that the monkey patch is removed as soon
    as it's not needed anymore, because it will break pip's normal behavior!
    """
    logger.debug("Cleaning up custom package finder ..")
    pip_install_module.PackageFinder = ORIGINAL_PACKAGE_FINDER

class CustomPackageFinder(pip_index_module.PackageFinder):

    """
    This class customizes :py:class:`pip.index.PackageFinder` to enforce what
    the ``--no-index`` option does for the default package index but doesn't do
    for package indexes registered with the ``--index=`` option in requirements
    files. Judging by pip's documentation the fact that this has to be monkey
    patched seems like a bug / oversight in pip (IMHO).
    """

    @property
    def index_urls(self):
        logger.debug("Custom package finder forcing --no-index behavior (hiding 'index_urls') ..")
        return []

    @index_urls.setter
    def index_urls(self, value):
        logger.debug("Custom package finder ignoring 'index_urls' value (%r) ..", value)

    @property
    def dependency_links(self):
        logger.debug("Custom package finder forcing --no-index behavior (hiding 'dependency_links') ..")
        return []

    @dependency_links.setter
    def dependency_links(self, value):
        logger.debug("Custom package finder ignoring 'dependency_links' value (%r) ..", value)

if __name__ == '__main__':
    main()



# Utility functions for the pip accelerator.
#
# Author: Peter Odding <peter.odding@paylogic.eu>
# Last Change: August 12, 2013
# URL: https://github.com/paylogic/pip-accel

# Standard library modules.
import os
import pwd
import re
import sys

# Look up the home directory of the effective user id so we can generate
# pathnames relative to the home directory.
HOME = pwd.getpwuid(os.getuid()).pw_dir

def expand_user(pathname):
    """
    Variant of :py:func:`os.path.expanduser()` that doesn't use ``$HOME`` but
    instead uses the home directory of the effective user id. This is basically
    a workaround for ``sudo -s`` not resetting ``$HOME``.

    :param pathname: A pathname that may start with ``~/``, indicating the path
                     should be interpreted as being relative to the home
                     directory of the current (effective) user.
    """
    return re.sub('^~(?=/)', HOME, pathname)

def get_python_version():
    """
    Return a string identifying the currently running Python version.

    :returns: A string like "py2.6" or "py2.7" containing a short mnemonic
              prefix followed by the major and minor version numbers.
    """
    return "py%i.%i" % (sys.version_info[0], sys.version_info[1])



# Extension of pip-accel that deals with dependencies on system packages.
#
# Author: Peter Odding <peter.odding@paylogic.eu>
# Last Change: May 11, 2014
# URL: https://github.com/paylogic/pip-accel

"""
System package dependency handling
==================================

Extension of the pip accelerator that deals with dependencies on system
packages. Currently only Debian Linux and derivative Linux distributions
are supported by this extension.
"""

# Standard library modules.
import os
import os.path

try:
    # Python 2.x.
    import ConfigParser as configparser
except ImportError:
    # Python 3.x.
    import configparser

# Internal modules.
from pip_accel.logger import logger

def sanity_check_dependencies(project_name, auto_install=None):
    """
    If ``pip-accel`` fails to build a binary distribution, it will call this
    function as a last chance to install missing dependencies. If this function
    does not raise an exception, ``pip-accel`` will retry the build once.

    :param project_name: The project name of a requirement as found on PyPI.
    :param auto_install: ``True`` if dependencies on system packages may be
                         automatically installed, ``False`` if missing system
                         packages should raise an error, ``None`` if the
                         decision should be based on the environment variable
                         ``PIP_ACCEL_AUTO_INSTALL``.

    If missing system packages are found, this function will try to install
    them. If anything "goes wrong" an exception is raised:

    - If ``auto_install=False`` then :py:class:`DependencyCheckFailed` is
      raised
    - If installation of missing packages fails
      :py:class:`DependencyInstallationFailed` is raised
    - If the user refuses to let pip-accel install missing packages
      :py:class:`RefusedAutomaticInstallation` is raised.

    If all goes well nothing is raised, nor is anything returned.
    """
    # Has the caller forbidden us from automatic installation?
    auto_install_forbidden = (auto_install == False)
    if auto_install is not None:
        # If the caller made an explicit choice, we'll respect that.
        auto_install_allowed = auto_install
    else:
        # Otherwise we check the environment variable.
        auto_install_allowed = 'PIP_ACCEL_AUTO_INSTALL' in os.environ
    logger.info("%s: Checking for missing dependencies ..", project_name)
    known_dependencies = current_platform.find_dependencies(project_name)
    if not known_dependencies:
        logger.info("%s: No known dependencies... Maybe you have a suggestion?", project_name)
    else:
        installed_packages = set(current_platform.find_installed())
        missing_packages = [p for p in known_dependencies if p not in installed_packages]
        if not missing_packages:
            logger.info("%s: All known dependencies are already installed.", project_name)
        elif auto_install_forbidden:
            msg = "Missing %i system package%s (%s) required by %s but automatic installation is disabled!"
            raise DependencyCheckFailed(msg % (len(missing_packages), '' if len(missing_packages) == 1 else 's',
                                               ', '.join(missing_packages), project_name))
        else:
            command_line = ['sudo'] + current_platform.install_command(missing_packages)
            if not auto_install_allowed:
                confirm_installation(project_name, missing_packages, command_line)
            logger.info("%s: Missing %i dependenc%s: %s",
                        project_name, len(missing_packages),
                        'y' if len(missing_packages) == 1 else 'ies',
                        " ".join(missing_packages))
            exit_code = os.spawnvp(os.P_WAIT, 'sudo', command_line)
            if exit_code == 0:
                logger.info("%s: Successfully installed %i missing dependenc%s.",
                            project_name, len(missing_packages),
                            len(missing_packages) == 1 and 'y' or 'ies')
            else:
                logger.error("%s: Failed to install %i missing dependenc%s!",
                             project_name, len(missing_packages),
                             len(missing_packages) == 1 and 'y' or 'ies')
                msg = "Failed to install %i system package%s required by %s! (command failed: %s)"
                raise DependencyInstallationFailed(msg % (len(missing_packages), '' if len(missing_packages) == 1 else 's',
                                                          project_name, command_line))

def confirm_installation(project_name, packages, command_line):
    """
    Notify the user that there are missing dependencies and how they can be
    installed. Then ask the user whether we are allowed to install the
    dependencies.

    :param packages: A list of strings with the names of the packages that are
                     missing.
    :param command_line: A list of strings with the command line needed to
                         install the packages.

    Raises :py:class:`RefusedAutomaticInstallation` when the user refuses to
    let pip-accel install any missing dependencies.
    """
    logger.info("%s: You seem to be missing %i dependenc%s: %s",  project_name,
                len(packages), len(packages) == 1 and 'y' or 'ies', " ".join(packages))
    logger.info("%s: I can install %s for you with this command: %s",
                project_name, len(packages) == 1 and 'it' or 'them',
                " ".join(command_line))
    try:
        prompt = "Do you want me to install %s dependenc%s? [y/N] "
        choice = raw_input(prompt % (len(packages) == 1 and 'this' or 'these',
                                     len(packages) == 1 and 'y' or 'ies'))
        if choice.lower().strip() == 'y':
            logger.info("Got permission to install missing dependenc%s.",
                        len(packages) == 1 and 'y' or 'ies')
            return
    except:
        pass
    logger.error("%s: Refused installation of missing dependenc%s!",
                 project_name, len(packages) == 1 and 'y' or 'ies')
    msg = "%s: User canceled automatic installation of missing dependenc%s!"
    raise RefusedAutomaticInstallation(msg % (project_name, 'y' if len(packages) == 1 else 'ies'))

class BasePlatform(object):

    """
    Base class for system package manager interfaces. Also implements the mock
    interface for platforms which are not (yet) supported by the pip
    accelerator.
    """

    @staticmethod
    def is_supported():
        """
        Check whether the current system supports the implemented package
        manager.
        """
        logger.debug("Falling back to dummy package manager interface ..")
        return True

    def __init__(self):
        """
        Load any predefined dependencies for the platform from a configuration
        file bundled with the pip accelerator.
        """
        config_name = getattr(self, 'config', '')
        if config_name:
            script = os.path.abspath(__file__)
            directory = os.path.dirname(script)
            config_path = os.path.join(directory, '%s.ini' % config_name)
            if os.path.isfile(config_path):
                logger.debug("Loading system package configuration from %s ..", config_path)
                parser = configparser.RawConfigParser()
                parser.read(config_path)
                if parser.has_section('dependencies'):
                    self.dependencies = dict((n.lower(), v.split()) for n, v in parser.items('dependencies'))

    def find_dependencies(self, project_name):
        """
        Find the system packages that should be installed to compile and run a
        known project from PyPI.

        :param project_name: The name of the project as given on PyPI.

        :returns: A list with the names of the system packages that should be
                  installed in order to compile and run the project. For
                  unknown projects an empty list is returned.
        """
        return self.dependencies.get(project_name.lower(), [])

    def find_installed(self):
        return []

class DebianLinux(BasePlatform):

    """
    Simple interface to the package management system of Debian Linux and
    derivative distributions like Ubuntu Linux.
    """

    config = 'debian'

    @staticmethod
    def is_supported():
        """
        Find out if we are running on a Debian (derived) system by checking if
        the file ``/etc/debian_version`` exists (this file is installed by the
        Debian system package ``base-files``).

        :returns: ``True`` if we are, ``False`` if we're not.
        """
        filename = '/etc/debian_version'
        if os.path.exists(filename):
            logger.debug("Looks like we are on a Debian (derived) system (%s exists) ..", filename)
            return True
        else:
            logger.debug("Looks like we're not on a Debian (derived) system (%s doesn't exist) ..", filename)
            return False

    def find_installed(self):
        """
        Find the packages that are installed on the current system.

        :returns: A list of strings with package names.
        """
        installed_packages = []
        handle = os.popen('dpkg -l')
        for line in handle:
            tokens = line.split()
            if len(tokens) >= 2 and tokens[0] == 'ii':
                installed_packages.append(tokens[1])
        return installed_packages

    def install_command(self, missing_packages):
        """
        Determine the command line needed to install the given package(s).

        :param missing_packages: A list with the names of one or more system
                                 packages to be installed.

        :returns: A list containing the program name and its arguments.
        """
        return ['apt-get', 'install', '--yes'] + missing_packages

class DependencyCheckFailed(Exception):
    """
    Custom exception type raised by :py:func:`sanity_check_dependencies()` when
    one or more known to be required system packages are missing and automatic
    installation of missing dependencies is explicitly disabled.
    """

class RefusedAutomaticInstallation(Exception):
    """
    Custom exception type raised by :py:func:`confirm_installation()` when the
    user refuses to install missing system packages.
    """

class DependencyInstallationFailed(Exception):
    """
    Custom exception type raised by :py:func:`sanity_check_dependencies()` when
    installation of missing system packages fails.
    """

# Select the interface to the package manager of the current platform.
for Interface in [DebianLinux, BasePlatform]:
    if Interface.is_supported():
        current_platform = Interface()
        break
